
module Common.THUtil where

import Control.Monad.Error () -- instance Monad (Either String)
import Data.Char
import Data.List
import Language.Haskell.THSyntax
import System.IO.Unsafe (unsafePerformIO)

import Common.Util

-- Escaping the quotation monad -------------------------------------------

escapeQ :: Q a -> a
escapeQ = unsafePerformIO . unQ

-- Declarations -----------------------------------------------------------

declaredName :: Dec -> String
declaredName (FunD s _)		    = s
declaredName (DataD _ s _ _ _)	    = s
declaredName (NewtypeD _ s _ _ _)   = s
declaredName (TySynD s _ _)	    = s
declaredName (ClassD _ s _ _)	    = s
declaredName (SigD s _)		    = s
declaredName d			    =
    error $ "declaredName: Don't know how to handle " ++ show d

-- Constructors -----------------------------------------------------------

conName :: Con -> String
conName (NormalC s _)	= s
conName (RecC s _)	= s
conName (InfixC _ s _)	= s

constructorArguments :: Con -> [Type]
constructorArguments (NormalC _ ts)	= map snd ts
constructorArguments (RecC _ vts)	= map pi3_3 vts
constructorArguments (InfixC t1 _ t2)	= map snd [t1,t2]

-- Types ------------------------------------------------------------------

isVar :: String -> Bool
isVar = isLower . head

varName :: Type -> ErrM String
varName (VarT x)    = return x
varName (ConT x)
    | isVar x	    = return x
varName t	    = fail $ "varName: Not a variable " ++ show t

typeName :: Type -> ErrM String
typeName (VarT x)
    | not (isVar x) = return x
typeName (ConT c)   = return c
typeName ListT	    = return "GHC.Base:[]"
typeName (TupleT 0) = return "GHC.Base:()"
typeName (TupleT n) = return $ "GHC.Base:(" ++ replicate (n-1) ',' ++ ")"
typeName t	    = fail $ "typeName: Doesn't have a name " ++ show t

isClosed :: Type -> Bool
isClosed = null . freeVars

freeVars :: Type -> [String]
freeVars (VarT x)	    = [x]
freeVars (ConT x) | isVar x = [x]
freeVars (AppT t1 t2)	    = freeVars t1 `union` freeVars t2
freeVars (ForallT xs _ t)   = freeVars t \\ xs
freeVars _		    = []

-- We assume unique names
substT :: [(String, Type)] -> Type -> Type
substT s t =
    case t of
	VarT x		    -> maybe t id $ lookup x s
	AppT t1 t2	    -> AppT (substT s t1) (substT s t2)
	ForallT vs ctx t    -> ForallT vs ctx (substT s t)
	_		    -> t

appWalkT t = id -*- reverse $ walk t
    where
	walk (AppT t1 t2)   = id -*- (t2:) $ walk t1
	walk t		    = (t, [])

-- Declaration environments -----------------------------------------------

type DecEnv = String -> ErrM Dec

mkDecEnv :: [Q Dec] -> DecEnv
mkDecEnv qds = \name -> maybe	(fail $ "lookupDec: Nothing known about " ++ name)
				return
				(lookup name env)
    where
	env = map (declaredName /\ id) $ escapeQ $ sequence qds

lookupDec = ($)

-- Expressions ------------------------------------------------------------

appWalkE e = id -*- reverse $ walk e
    where
	walk (AppE e1 e2)   = id -*- (e2:) $ walk e1
	walk e		    = (e, [])

-- | Reduce top level beta redexes
beta :: Exp -> Exp
beta e = uncurry (beta' []) $ appWalkE e
    where
	beta' s (LamE [VarP x] body) (a:as) =
	    beta' ((x,a):s) body as
	beta' s e es = foldl AppE (substE s e) es

-- | Very simple substitution. Only handles applications.
substE :: [(String, Exp)] -> Exp -> Exp
substE s e =
    case e of
	VarE x		-> maybe e id $ lookup x s
	e1 `AppE` e2	-> substE s e1 `AppE` substE s e2
	_		-> error $ "substE: " ++ show (pprExp e)

