{-# OPTIONS -fth #-}

module GenericHaskell.GH where

import Control.Monad.State
import Language.Haskell.THSyntax
import qualified Text.Read as Read
import qualified Text.ParserCombinators.ReadP as ReadP

import Common.EP
import Common.Util
import Common.THUtil
import Common.ZCoding

-- Datatype codes ---------------------------------------------------------

type Name	= String
type TypeName	= Name
type VarName	= Name
type ConName	= Name
type LabelName	= Name
type ClassName	= Name

data Fixity	= Infix | Prefix

data ConDescr	= ConDescr	{ cName   :: ConName
				, cFixity :: Fixity
				}
data LabelDescr	= LabelDescr	{ lName :: LabelName
				}

data Code   = Sum | Prod | Unit
	    | Constr ConDescr | Label LabelDescr
	    | Fun | TypeCon TypeName
	    | App Code Code
	    | Lam VarName Code
	    | Var VarName

type Datatype = (TypeName, Kind, Code)

f |@| g	    = App f g
f |+| g	    = Sum |@| f |@| g
f |*| g	    = Prod |@| f |@| g
con c f	    = Constr c |@| f
label l f   = Label l |@| f
f |->| g    = Fun |@| f |@| g

infixr 4 |+|
infixr 5 |*|
infixl 6 |->|
infixl 7 |@|

-- Pretty printing --------------------------------------------------------

instance Show ConDescr where
    show = cName

instance Show LabelDescr where
    show = lName

instance Show Kind where
    showsPrec p Star	= showString "*"
    showsPrec p (FunK k1 k2) =
	showParen (p>0) $
	    showsPrec 1 k1 . showString " -> " . showsPrec 0 k2

instance Read Kind where
    readPrec = Read.choice [ Read.prec 1 funP, brackets funP, Read.lift starP ]
	where
	    brackets p = do Read.lift ReadP.skipSpaces
			    Read.lift $ ReadP.string "("
			    Read.lift ReadP.skipSpaces
			    x <- p
			    Read.lift ReadP.skipSpaces
			    Read.lift $ ReadP.string ")"
			    Read.lift ReadP.skipSpaces
			    return x
	    starP = do	ReadP.skipSpaces
			ReadP.string "*"
			ReadP.skipSpaces
			return Star
	    funP =  do	Read.lift ReadP.skipSpaces
			k <- Read.step Read.readPrec
			Read.lift ReadP.skipSpaces
			Read.lift $ ReadP.string "->"
			Read.lift ReadP.skipSpaces
			l <- Read.readPrec
			Read.lift ReadP.skipSpaces
			return $ FunK k l

instance Show Code where
    showsPrec p c = case c of
	App (App Sum c1) c2 ->
	    showParen (p>3) $
		showsPrec 4 c1 . showString " + " . showsPrec 3 c2
	App (App Prod c1) c2 ->
	    showParen (p>5) $
		showsPrec 6 c1 . showString " * " . showsPrec 5 c2
	App (App Fun c1) c2 ->
	    showParen (p>7) $
		showsPrec 8 c1 . showString " -> " . showsPrec 7 c2
	App c1 c2 ->
	    showParen (p>9) $
		showsPrec 9 c1 . showString " " . showsPrec 10 c2
	Sum	    -> showString "(+)"
	Prod	    -> showString "(*)"
	Unit	    -> showString "1"
	Constr cd   -> showString "{" . shows cd . showString "}"
	Label ld    -> shows ld
	Fun	    -> showString "(->)"
	TypeCon t   -> showString t
	Lam x c	    ->
	    showParen (p>0) $
		showString ("/\\" ++ x ++ ". ") . shows c
	Var x	    -> showString x

-- Creating the codes -----------------------------------------------------

typeCode :: Q Dec -> Code
typeCode qd = mkCode d
    where
	d	= escapeQ qd
	bad s	= error $ "typeCode: " ++ s ++ "\n"
		++ show (pprDec d)

	mkCode (DataD _ _ args cons _)	    = foldr Lam ( foldr1 (|+|)
							$ map mkCon cons
							) args
	mkCode (NewtypeD _ _ args con _)    = foldr Lam ( mkCon con ) args
	mkCode _			    = bad "Not a datatype declaration"

	mkCon (NormalC name ts) =
	    con (ConDescr name Prefix) $ mkProd $ map (mkT . snd) ts
	mkCon (RecC name vts) =
	    con (ConDescr name Prefix) $ mkProd $ map (mkRec . pi13_3) vts
	mkCon (InfixC t1 name t2) =
	    con (ConDescr name Infix) $ mkProd $ map (mkT . snd) [t1,t2]

	mkProd []   = Unit
	mkProd cs   = foldr1 (|*|) cs

	mkRec (lbl, t) = label (LabelDescr lbl) $ mkT t

	mkT (ForallT _ _ _) = bad "Universal quantification not supported"
	mkT (VarT x)	    = Var x
	mkT (ConT x)
	    | isVar x	    = Var x	    -- Template Haskell bug
	    | otherwise	    = TypeCon x
	mkT (AppT t1 t2)    = mkT t1 |@| mkT t2
	mkT ArrowT	    = Fun
	mkT t		    = TypeCon $ runErr $ typeName t

datatype :: Kind -> Q Dec -> Datatype
datatype k d = (name, k, typeCode d)
    where
	name = declaredName $ escapeQ d

datatype' :: String -> Q Dec -> Datatype
datatype' s = datatype (read s)

-- Structure types --------------------------------------------------------

{-
structureType :: Code -> TypeAbs
structureType t args =
    case t of
	Var x		    -> VarT x
	App (Constr _) t    -> structureType t args
	App (Label _) t	    -> structureType t args
	App t1 t2	    -> AppT (structureType t1 args) (structureType t2 args)
	Sum		    -> ConT "Data.Either:Either"
	Prod		    -> TupleT 2
	Unit		    -> TupleT 0
	Fun		    -> ArrowT
	TypeCon con	    -> ConT con
	Lam v t		    ->
	    case args of
		a:as	-> substT [(v,a)] $ structureType t as
		_	-> error $ "structureType: Too few arguments to " ++ show t
-}

-- Generic Function Definitions -------------------------------------------

type GenericFun		= Code -> Q Exp
type GEnv		= [(VarName, Q Exp)]
type GenericFun'	= GEnv -> GenericFun
type KindIndexedType	= Kind -> TypeAbs
type GenericType	= Kind -> Type -> Type

data Kind	= Star | FunK Kind Kind
type TypeAbs	= [Type] -> Type

infixl 9 @@
(@@) = AppT
t1 --> t2 = ArrowT @@ t1 @@ t2

kindIndexedType :: Int		    -- # generic arguments
		-> Int		    -- # non-generic arguments
		-> TypeAbs	    -- type for kind *
		-> Kind -> TypeAbs
kindIndexedType g ng tstar k args
    | length args /= g + ng = error $ "kindIndexedType: Expected "
				++ show g ++ " + " ++ show ng ++ " arguments. Got "
				++ show (length args)
    | otherwise		    = evalState (kit k args) genericNamess
    where
	genericNamess	= [ [ v ++ show n | n <- [1 .. g] ] | v <- tail names ]
	    where
		names = "" : [ s ++ [c] | s <- names, c <- chars ]
		chars = "tuvw"
	nonGenericVars = [ VarT $ "a" ++ show n | n <- [1 .. ng] ]

	freshVars = do	x:xs <- get
			put xs
			return x

	kit Star args		= return $ tstar args
	kit (FunK k1 k2) args	=
	    do	us  <- freshVars
		let vs = map VarT us
		t1  <- kit k1 (vs ++ nonGenericVars)
		t2  <- kit k2 (zipWith AppT args vs ++ nonGenericVars)
		return $ ForallT us [] $ t1 --> t2

gName name c	= name ++ "__d" ++ zEncode c
sName name c	= name ++ "__s" ++ zEncode c
epName name	= "ep__" ++ zEncode name

defaultTrans :: VarName -> GenericFun' -> GenericFun'
defaultTrans name gfun env t =
    case t of
	Constr _    -> [| id |]
	Label _	    -> [| id |]
	TypeCon c   -> varE $ gName name c
	App s t	    -> [| $(gfun env s) $(gfun env t) |]
	Lam x t	    -> [| \gx -> $(gfun ((x,[|gx|]):env) t) |]
	Var x	    -> maybe	(error $ "Unbound var: " ++ x)
				id
				(lookup x env)

-- From structure types to datatypes --------------------------------------

structEP :: Datatype -> Q Dec
structEP d@(name, _, _) =
    funD    (epName name)
	    [ clause []
		(normalB $ [| EP |] `appE` varE "out" `appE` varE "inn")
		[ funD "out" [ clause [] (normalB $ out d) [] ]
		, funD "inn" [ clause [] (normalB $ inn d) [] ]
		] 
	    ]

constructorArities :: Code -> [(ConName, Int)]
constructorArities t =
    case t of
	Sum `App` (Constr con `App` t1) `App` t2 ->
	    (cName con, arity t1) : constructorArities t2
	Constr con `App` t ->
	    [ (cName con, arity t) ]
	Lam _ t -> constructorArities t
    where
	arity Unit			= 0
	arity (Prod `App` _ `App` t)	= 1 + arity t
	arity _				= 1

inn, out :: Datatype -> Q Exp

inn (_,_,t) =
	lamE [varP "x"] $
	caseE (varE "x") matches
    where
	nofCons	= length matches
	matches = zipWith mkMatch [0..] $ constructorArities t

	leftAndRights 0 1 p = p
	leftAndRights 0 n p = conP "Data.Either:Left" [p]
	leftAndRights i n p = conP "Data.Either:Right" [leftAndRights (i-1) (n-1) p]

	tupleP []   = tupP []
	tupleP ps   = foldr1 (\x y -> tupP [x,y]) ps

	mkMatch i (name, ar) =
	    do	xs <- replicateM ar (gensym "x")
		match (pat xs) (normalB $ body xs) []
	    where
		pat xs	= leftAndRights i nofCons (tupleP $ map varP xs)
		body xs	= foldl appE (conE name) $ map varE xs

out (_,_,t) =
	lamE [varP "x"] $
	caseE (varE "x") matches
    where
	nofCons	= length matches
	matches = zipWith mkMatch [0..] $ constructorArities t

	leftAndRights 0 1 e = e
	leftAndRights 0 n e = conE "Data.Either:Left" `appE` e
	leftAndRights i n e = conE "Data.Either:Right" `appE` leftAndRights (i-1) (n-1) e

	tupleE []   = tupE []
	tupleE es   = foldr1 (\x y -> tupE [x, y]) es

	mkMatch i (name, ar) =
	    do	xs <- replicateM ar (gensym "x")
		match (pat xs) (normalB $ body xs) []
	    where
		pat xs	= conP name $ map varP xs
		body xs	= leftAndRights i nofCons (tupleE $ map varE xs)

typeEP :: Q Exp -> Kind -> GenericType -> Q Exp
typeEP epV k t = mkEP k $ t k (VarT "_")
    where

	mkEP (FunK _ k) (ForallT _ _ (ArrowT `AppT` _ `AppT` t)) = mkEP k t
	mkEP Star t = epApp t

	epApp t = case appWalkT t of
	    (VarT "_", ts)  -> epV
	    (VarT v, ts)    -> [| epId |]
	    (t', ts)	    -> foldl appE (epT t') (map epApp ts)

	epT ArrowT		    = [| epFun |]
	epT (VarT x)
	    | isVar x		    = [| epId |]
	epT (ConT "Common.EP:EP")   = [| epEP |]
	epT (ForallT _ _ t)	    = epApp t
	epT t = flip (either error) (typeName t)
		    $ varE . gName "ep"

-- Instantiation ----------------------------------------------------------

type Generic = (VarName, GenericType, GenericFun)

instantiate :: Generic -> Datatype -> Q [Dec]
instantiate (gname, gtype, gfun) (tname, kind, code) =
    sequence
	[ sigD nameD $ return $ gtype kind $ ConT tname
	, do	args <- replicateM (arity kind) $ gensym "x"
		funD
		    nameD
		    [ clause (map varP args)
			(normalB $ foldl appE [| to |]
			    [ typeEP (varE $ epName tname) kind gtype
			    , do    e <- foldl appE (gfun code) $ map varE args
				    return $ beta e
			    ]
			) []
		    ]
	]
    where
	nameD = gName gname tname
	nameS = sName gname tname

	arity Star	    = 0
	arity (FunK _ k)    = 1 + arity k

structure :: Datatype -> Q [Dec]
structure = liftM (:[]) . structEP

