
module PolyP.PolyP where

import Control.Monad
import Language.Haskell.THSyntax

import Common.Util
import Common.THUtil

-- Datatype codes ---------------------------------------------------------

type ConName	= String
type TypeName	= String
type VarName	= String

data Code   = Code :+: Code
	    | Code :*: Code
	    | Unit
	    | Par
	    | Rec
	    | Regular :@: Code
	    | Const Type

type Regular	= ([ConName], Code)

instance Show Code where
    showsPrec p f = case f of
	g :+: h	    -> showParen (p>3) $
			showsPrec 4 g . showString " + " . showsPrec 3 h
	g :*: h	    -> showParen (p>5) $
			showsPrec 6 g . showString " * " . showsPrec 5 h
	Unit	    -> showString "1"
	Par	    -> showString "Par"
	Rec	    -> showString "Rec"
	(_,d) :@: g -> showParen (p>9) $
			showsPrec 9 d . showString " " . showsPrec 10 g
	Const t	    -> showsPrec p t

-- Creating the codes -----------------------------------------------------

functorOf :: Regular -> Code
functorOf = snd

regular :: [Q Dec] -> TypeName -> Regular
regular qds name = either error id $ env name >>= mkRegular 
    where
	env = mkDecEnv qds

	mkRegular :: Dec -> ErrM Regular
	mkRegular d = case d of

		DataD _ _ [a] cons _	->
		    do	code <- mkSum a cons
			return (map conName cons, code)

		NewtypeD _ _ [a] con _	->
		    do	code <- mkProd a con
			return ([conName con], code)

		_			-> notRegular
	    where

		myName = declaredName d

		notRegular	= fail $ "regular: Not a regular datatype:\n" ++ show (pprDec d)
		undefinedVar x	= fail $ "regular: Undefined type variable " ++ x ++ " in\n" ++ show (pprDec d)

		mkSum :: VarName -> [Con] -> ErrM Code
		mkSum arg []	= notRegular
		mkSum arg cons	= liftM (foldr1 (:+:)) $ mapM (mkProd arg) cons

		mkProd :: VarName -> Con -> ErrM Code
		mkProd arg con   = liftM foldProd
				    $ mapM (mkType arg)
				    $ constructorArguments con
		    where
			foldProd [] = Unit
			foldProd cs = foldr1 (:*:) cs

		mkType :: VarName -> Type -> ErrM Code
		mkType arg t = foldr1 mplus

		    [	-- Is it a variable?
			do  x <- varName t
			    if arg == x
				then return Par
				else undefinedVar x

		    ,	-- Is it an application?
			case t of
			    AppT t1 t2  ->
				do  name <- typeName t1

				    -- Is it a recursive call?
				    if name == myName
					-- then it should be a Rec
					then do a <- varName t2
						if a == arg
						    then return Rec
						    else notRegular
					-- else it might be a (:@:)
					else do d   <- env name
						r   <- mkRegular d
						c   <- mkType arg t2
						return $ r :@: c

		    ,	-- Otherwise it better be closed
			if isClosed t
			    then return $ Const t
			    else notRegular
		    ]

-- From Structure Types to Datatypes --------------------------------------

constructorArities :: Regular -> [(ConName, Int)]
constructorArities d = case d of
	(con:[], c)		-> [(con, arity c)]
	(con:cons, c :+: c')	->
	    (con, arity c) : constructorArities (cons, c')
    where
	arity Unit	= 0
	arity (_ :*: c) = 1 + arity c
	arity _		= 1

inn, out :: Regular -> Q Exp

inn d =	lamE [varP "x"] $
	caseE (varE "x") matches
    where
	nofCons	= length $ fst d
	matches = zipWith mkMatch [0..] $ constructorArities d

	leftAndRights 0 1 p = p
	leftAndRights 0 n p = conP "Data.Either:Left" [p]
	leftAndRights i n p = conP "Data.Either:Right" [leftAndRights (i-1) (n-1) p]

	tupleP []   = tupP []
	tupleP ps   = foldr1 (\x y -> tupP [x,y]) ps

	mkMatch i (name, ar) =
	    do	xs <- replicateM ar (gensym "x")
		match (pat xs) (normalB $ body xs) []
	    where
		pat xs	= leftAndRights i nofCons (tupleP $ map varP xs)
		body xs	= foldl appE (conE name) $ map varE xs

out d =	lamE [varP "x"] $
	caseE (varE "x") matches
    where
	nofCons	= length $ fst d
	matches = zipWith mkMatch [0..] $ constructorArities d

	leftAndRights 0 1 e = e
	leftAndRights 0 n e = conE "Data.Either:Left" `appE` e
	leftAndRights i n e = conE "Data.Either:Right" `appE` leftAndRights (i-1) (n-1) e

	tupleE []   = tupE []
	tupleE es   = foldr1 (\x y -> tupE [x, y]) es

	mkMatch i (name, ar) =
	    do	xs <- replicateM ar (gensym "x")
		match (pat xs) (normalB $ body xs) []
	    where
		pat xs	= conP name $ map varP xs
		body xs	= leftAndRights i nofCons (tupleE $ map varE xs)

